#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
from collections import Mapping, Container, UserDict

from dotty_dict.dotty_dict import Dotty


class TestDotty_dict(unittest.TestCase):

    def setUp(self):
        self.dotty = Dotty({
            'already': {
                'exist': {
                    'deep': 'key',
                },
            },
            'flat_key': 'i am here',
        })

    def tearDown(self):
        pass

    def test_dotty_type(self):
        # self.assertIsInstance(self.dotty, dict)
        self.assertIsInstance(self.dotty, Mapping)
        self.assertIsInstance(self.dotty, UserDict)
        self.assertIsInstance(self.dotty, Dotty)

    def test_key_split(self):
        key = 'super.deeply.nested.key'
        tree = self.dotty._split(key)

        self.assertEqual(len(tree), 4)
        self.assertEqual(tree[1], 'deeply')

    def test_key_single_leaf(self):
        self.dotty['single'] = 'test'

        self.assertEqual(self.dotty['single'], 'test')
        self.assertEqual(self.dotty.get('single'), 'test')

    def test_key_nested_leaf(self):
        self.dotty['single.nested'] = 'test'

        self.assertEqual(self.dotty['single']['nested'], 'test')

    def test_key_nested_leaf_update(self):
        self.dotty['already.exist.new'] = 'value'

        self.assertEqual(self.dotty['already']['exist']['deep'], 'key')
        self.assertEqual(self.dotty['already']['exist']['new'], 'value')

    def test_key_nested_leaf_replace(self):
        self.dotty['already.exist'] = {'new': 'value'}

        self.assertNotIn('deep', self.dotty['already']['exist'])
        self.assertEqual(self.dotty['already']['exist']['new'], 'value')

    def test_access_nested_value(self):
        deep1 = self.dotty['already.exist.deep']
        deep2 = self.dotty['already']['exist']['deep']

        self.assertEqual(deep1, deep2)

    def test_access_nested_value_not_exist(self):
        self.assertIsNone(self.dotty['already.exist.black_hole'])

    def test_work_with_not_dotty_dict_like_objects(self):
        # Assume dict-like objects does not support dot notation
        self.dotty['already.exist.custom'] = UserDict({
            'not_dotty': {
                'user_dict': 'I am deeply nested user dict value'
            }
        })

        # Accessing not existing key from dict-like object should raise KeyError
        self.assertRaises(KeyError, self.dotty['already.exist.custom.key_value'])

        # One can access not Dotty dict-like object with dot notation
        self.assertEqual(self.dotty['already.exist.custom.not_dotty.user_dict'],
                         'I am deeply nested user dict value')

        # one can add new key to not dotty dict-like object with dot notation
        self.dotty['already.exist.custom.not_dotty.new_key'] = 'new key inside not dotty dict'

        self.assertEqual(
            self.dotty['already.exist.custom'],
            {
                'not_dotty': {
                    'user_dict': 'I am deeply nested user dict value',
                    'new_key': 'new key inside not dotty dict',
                }
            }
        )

    def test_get_value_simple(self):
        self.assertEqual(self.dotty.get('flat_key'), 'i am here')

    def test_get_value_nested(self):
        self.assertEqual(self.dotty.get('already.exist.deep'), 'key')

    def test_get_default_value_simple(self):
        self.assertEqual(self.dotty.get('black_hole', 'default value'), 'default value')

    def test_get_default_value_nested(self):
        self.assertEqual(self.dotty.get('already.exist.black_hole', 'def value'), 'def value')




# -*- coding: utf-8 -*-
"""Dotty_Dict Module."""

from collections import (Mapping, UserDict)


class Dotty(UserDict):
    """Dotty dict-like object support dot notation keys.
    
    Dotty dict-like object allow to access deeply nested keys using dot notation.
    Create Dotty from dict or other dict-like object to use magic of Dotty.
    
    :example:
    <code>
        # Create dotty dict-like object.
        dotty = Dotty({
            'deep_key': {
                'nested': 'nested value',
            },
        })
        
        # Assign value to very deeply nested key not existing yet.
        dotty['deep_key.new_nested.very.deep.key'] = 'wow!'
        
        # Old keys still there
        assert dotty['deep_key.nested'] == 'nested value'
        
        # Access new deeply nested key
        assert dotty['deep_key.new_nested.very.deep.key'] == 'wow!'
        
        # Old fashion way to get very deeply nested key
        assert dotty['deep_key']['new_nested']['very']['deep']['key'] == 'wow!'
        
        # Get method with provided default value works as expected
        assert dotty.get('deep_key.new_nested.something', 'default value') == 'default value'
    </code>
    """

    def __getitem__(self, key):
        """Get value from deeply nested key.
        
        If key does not exist return None instead of raising KeyError exception.

        :param key:
        :return:
        """
        tree = self._split(key)

        item = dict(self.data)
        for leaf in tree:
            if leaf in item:
                item = item[leaf]
            else:
                item = None
                break

        return item

    def __setitem__(self, key, value):
        """
        Set recursively new key:value item into Dotty dict. Allow to set deeply nested keys.

        :param (str) key:
        :param (any) value:
        :return:
        """
        def insert_into_dict(dictionary, leaf, val):
            """
            Recursively insert new keys and value at last to Dotty dictionary.
            
            :param (dict|Dotty) dictionary:
            :param (str) leaf:
            :param (any) val:
            :return: New dictionary
            :rtype: dict
            """
            if len(tree) > 0:
                """
                Create or update deeply nested dictionaries if there is more leaf to follow.
                """
                if leaf in dictionary and isinstance(dictionary[leaf], dict):
                    """
                    If dictionary leaf is instance of dict, convert it to Dotty dict
                    then update instead of override.
                    """
                    dictionary[leaf] = Dotty(dictionary[leaf])
                    dictionary[leaf].update(
                        insert_into_dict(dictionary[leaf], tree.pop(0), val)
                    )

                elif leaf in dictionary and isinstance(dictionary[leaf], Mapping):
                    """
                    If dictionary leaf is instance of dict-like object, update it
                    instead of override.
                    """
                    dictionary[leaf].update(
                        insert_into_dict(dictionary[leaf], tree.pop(0), val)
                    )

                else:
                    """
                    If dictionary leaf is not a dict nor dict-like object,
                    override val with new Dotty dict.
                    """
                    dictionary[leaf] = insert_into_dict(
                        Dotty(dictionary), tree.pop(0), val
                    )

                return dictionary

            """
            Assign val to last leaf in tree then return dictionary.
            """
            dictionary[leaf] = val

            return dictionary

        tree = self._split(key)
        self.data = insert_into_dict(dict(self.data), tree.pop(0), value)

    @staticmethod
    def _split(key):
        """
        Split dot notation key string to leafs.
        
        :param (str) key: Dot notated key.
        :return: List of leafs.
        """
        return key.split('.')

    def get(self, key, default=None):
        """
        Get dictionary value if exists otherwise get default value.
        
        :param (str) key: Dot notated key.
        :param (any) default: Default value if value does not exist.
        :return:
        """
        return self.__getitem__(key) or default



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = []
test_requirements = [
    'coverage==4.3.4',
    'mock>=1.0.1',
    'flake8==3.3.0',
    'tox==2.7.0',
    'codecov>=2.0.0',
]

setup(
    name='dotty_dict',
    version='0.1.8',
    description="Dotty dict-like object allow to access deeply nested keys using dot notation.",
    long_description=readme + '\n\n' + history,
    author="Paweł Zadrożny",
    author_email='pawel.zny@gmail.com',
    url='https://github.com/pawelzny/dotty_dict',
    packages=[
        'dotty_dict',
    ],
    package_dir={'dotty_dict': 'dotty_dict'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='dotty_dict',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
